@@ -1,5 +1,6 @@
 //! Application configuration parsing and validation.
 
+use crate::updates::UpdateFrequency;
 use directories::UserDirs;
 use patricia_tree::PatriciaSet;
 use std::collections::HashMap;
@@ -60,6 +61,9 @@ pub(crate) struct Config {
     /// User's diff color choices.
     diff_colors: (egui::Color32, egui::Color32),
 
+    /// User's update check frequency choice.
+    update_check: UpdateFrequency,
+
     /// Map raw track IDs to unique track IDs.
     track_ids: PatriciaSet,
 
@@ -115,6 +119,7 @@ impl Config {
             theme: UserTheme::Auto,
             colors: Vec::new(),
             diff_colors: (egui::Color32::TRANSPARENT, egui::Color32::TRANSPARENT),
+            update_check: UpdateFrequency::default(),
             track_ids: PatriciaSet::new(),
             tracks: HashMap::new(),
             cars: HashMap::new(),
@@ -154,17 +159,31 @@ impl Config {
         let doc: Document = fs::read_to_string(&doc_path)?.parse()?;
 
         let setups_path = PathBuf::from(
-            doc["config"]["setups_path"]
-                .as_str()
+            doc.get("config")
+                .and_then(|t| t.get("setups_path"))
+                .and_then(|t| t.as_str())
                 .ok_or_else(|| Error::type_error("config.setups_path", "string"))?,
         );
 
-        let theme = UserTheme::from_item(&doc["config"]["theme"]);
+        let theme = doc
+            .get("config")
+            .and_then(|t| t.get("theme"))
+            .and_then(|t| t.as_str())
+            .unwrap_or("auto");
+        let theme = UserTheme::from_str(theme);
+
+        let update_check = doc
+            .get("config")
+            .and_then(|t| t.get("update_check"))
+            .and_then(|t| t.as_str())
+            .map(UpdateFrequency::from)
+            .unwrap_or_default();
 
         let mut config = Self::new(doc_path, min_size);
         config.doc = doc;
-        config.theme = theme;
         config.update_setups_path(setups_path);
+        config.update_theme(theme);
+        config.set_update_check(update_check);
         config.load_tracks_and_cars()?;
         config.load_colors()?;
 
@@ -187,14 +206,14 @@ impl Config {
 
     /// Get window configuration if it's valid.
     pub(crate) fn get_window(&self) -> Option<Window> {
-        let window = &self.doc["window"];
+        let window = &self.doc.get("window")?;
 
-        let x = window["x"].as_integer()?;
-        let y = window["y"].as_integer()?;
+        let x = window.get("x").and_then(|t| t.as_integer())?;
+        let y = window.get("y").and_then(|t| t.as_integer())?;
         let position = PhysicalPosition::new(x as i32, y as i32);
 
-        let width = window["width"].as_integer()?;
-        let height = window["height"].as_integer()?;
+        let width = window.get("width").and_then(|t| t.as_integer())?;
+        let height = window.get("height").and_then(|t| t.as_integer())?;
         let size = PhysicalSize::new(
             (width as u32).max(self.min_size.width),
             (height as u32).max(self.min_size.height),
@@ -215,7 +234,10 @@ impl Config {
 
     /// Update the setup exports path.
     pub(crate) fn update_setups_path<P: AsRef<Path>>(&mut self, setups_path: P) {
-        self.setups_path = setups_path.as_ref().to_path_buf();
+        self.setups_path = setups_path
+            .as_ref()
+            .canonicalize()
+            .unwrap_or_else(|_| setups_path.as_ref().to_path_buf());
 
         // Note that to_string_lossy() is destructive when the path contains invalid UTF-8 sequences.
         // If this is a problem in practice, we _could_ write unencodable paths as an array of
@@ -258,7 +280,7 @@ impl Config {
     }
 
     /// Modify user's color-coding choices.
-    pub(crate) fn mut_colors(&mut self) -> &mut Vec<egui::Color32> {
+    pub(crate) fn colors_mut(&mut self) -> &mut Vec<egui::Color32> {
         &mut self.colors
     }
 
@@ -293,28 +315,39 @@ impl Config {
     }
 
     /// Modify user's diff color choices.
-    pub(crate) fn mut_diff_colors(&mut self) -> &mut (egui::Color32, egui::Color32) {
+    pub(crate) fn diff_colors_mut(&mut self) -> &mut (egui::Color32, egui::Color32) {
         &mut self.diff_colors
     }
 
+    /// Update the frequency for update checks.
+    pub(crate) fn get_update_check(&self) -> UpdateFrequency {
+        self.update_check
+    }
+
+    /// Update the frequency for update checks.
+    pub(crate) fn set_update_check(&mut self, update_check: UpdateFrequency) {
+        self.update_check = update_check;
+        self.doc["config"]["update_check"] = toml_edit::value(self.update_check.as_str());
+    }
+
     /// Load track and car info from config.
     fn load_tracks_and_cars(&mut self) -> Result<(), Error> {
-        let table = &self.doc["tracks"];
-        if let Some(tracks) = table.as_table() {
+        let table = &self.doc.get("tracks").and_then(|t| t.as_table());
+        if let Some(tracks) = table {
             for (id, name) in tracks.iter() {
                 let name = name
                     .as_str()
                     .ok_or_else(|| Error::type_error(&format!("tracks.{}", id), "string"))?;
 
-                self.track_ids.insert(id.to_string());
+                self.track_ids.insert(id);
                 self.tracks.insert(id.to_string(), name.to_string());
             }
         } else if !table.is_none() {
             return Err(Error::type_error("tracks", "table"));
         }
 
-        let cars = &self.doc["cars"];
-        if let Some(cars) = cars.as_table() {
+        let cars = &self.doc.get("cars").and_then(|t| t.as_table());
+        if let Some(cars) = cars {
             for (id, name) in cars.iter() {
                 let name = name
                     .as_str()
@@ -332,8 +365,13 @@ impl Config {
     /// Load column colors and background colors from config.
     fn load_colors(&mut self) -> Result<(), Error> {
         let mut parsed = Vec::new();
-        let colors = &self.doc["config"]["colors"];
-        if let Some(colors) = colors.as_array() {
+        let colors = &self
+            .doc
+            .get("config")
+            .and_then(|t| t.get("colors"))
+            .and_then(|t| t.as_array());
+
+        if let Some(colors) = colors {
             for (i, color) in colors.iter().enumerate() {
                 let color = color
                     .as_str()
@@ -350,7 +388,12 @@ impl Config {
         // Parse background colors
         let mut background = Vec::new();
         for name in &["background_decrease", "background_increase"] {
-            let color = self.doc["config"][name].as_str();
+            let color = self
+                .doc
+                .get("config")
+                .and_then(|t| t.get(name))
+                .and_then(|t| t.as_str());
+
             if let Some(color) = color {
                 let color =
                     color_from_str(color).map_err(|_| Error::Color(format!("config.{}", name)))?;
@@ -404,16 +447,13 @@ impl Window {
 }
 
 impl UserTheme {
-    /// Create a `UserTheme` from a TOML item.
-    fn from_item(value: &Item) -> Self {
-        value
-            .as_str()
-            .map(|value| match value {
-                "dark" => Self::Dark,
-                "light" => Self::Light,
-                _ => Self::Auto,
-            })
-            .unwrap_or(Self::Auto)
+    /// Create a `UserTheme` from a string slice.
+    fn from_str(value: &str) -> Self {
+        match value {
+            "dark" => Self::Dark,
+            "light" => Self::Light,
+            _ => Self::Auto,
+        }
     }
 
     /// Get a string slice that is TOML-compatible for this `UserTheme`.
@@ -489,7 +529,7 @@ mod tests {
         let track_ids = &config.track_ids;
         assert_eq!(
             track_ids.get_longest_common_prefix("charlotte_2018_2019_roval"),
-            Some("charlotte_2018_2019_roval".as_bytes()),
+            Some("charlotte_2018".as_bytes()),
         );
         assert_eq!(
             track_ids.get_longest_common_prefix("charlotte_fullroadcourse"),
@@ -499,12 +539,12 @@ mod tests {
 
         // Expectations for track name mapping.
         assert_eq!(
-            config.tracks.get("charlotte_2018_2019_roval"),
-            Some(&"Charlotte Motor Speedway - Road Course".to_string())
+            config.tracks.get("charlotte_2018"),
+            Some(&"Charlotte Motor Speedway".to_string())
         );
         assert_eq!(
             config.tracks.get("charlotte"),
-            Some(&"[Legacy] Charlotte Motor Speedway - 2008".to_string())
+            Some(&"[Legacy] Charlotte Motor Speedway".to_string())
         );
 
         // Expectations for car name mapping.
