@@ -2,16 +2,14 @@
 
 use crate::config::{Config, Error as ConfigError, UserTheme};
 use crate::gpu::Gpu;
-use crate::gui::{ErrorButton, Gui, ShowError};
+use crate::gui::{ErrorButton, Gui, ShowError, ShowWarning};
+use crate::updates::{UpdateChecker, UpdateNotification};
 use directories::ProjectDirs;
-use egui::ClippedMesh;
+use egui::{ClippedMesh, CtxRef};
 use egui_wgpu_backend::{BackendError, RenderPass, ScreenDescriptor};
-use egui_winit_platform::{Platform, PlatformDescriptor};
 use font_loader::system_fonts::{self, FontPropertyBuilder};
-use std::borrow::Cow;
 use std::collections::VecDeque;
 use std::path::PathBuf;
-use std::time::Instant;
 use thiserror::Error;
 use winit::dpi::PhysicalSize;
 use winit::event_loop::EventLoopProxy;
@@ -20,13 +18,15 @@ use winit::window::{Theme, Window};
 /// Manages all state required for rendering egui.
 pub(crate) struct Framework {
     // State for egui.
-    start_time: Instant,
-    platform: Platform,
+    egui_ctx: CtxRef,
+    egui_state: egui_winit::State,
     screen_descriptor: ScreenDescriptor,
     rpass: RenderPass,
     paint_jobs: Vec<ClippedMesh>,
     theme: Option<Theme>,
     gui: Gui,
+    update_checker: Option<UpdateChecker>,
+    event_loop_proxy: EventLoopProxy<UserEvent>,
 }
 
 /// Framework errors.
@@ -37,7 +37,7 @@ pub(crate) enum Error {
 }
 
 /// User event handling is performed with this type.
-#[derive(Debug, Eq, PartialEq)]
+#[derive(Debug)]
 pub(crate) enum UserEvent {
     /// Configuration error handling events
     ConfigHandler(ConfigHandler),
@@ -48,8 +48,17 @@ pub(crate) enum UserEvent {
     /// Change the path for setup export files.
     SetupPath(Option<PathBuf>),
 
+    /// File system event for the setup export path.
+    FsChange(hotwatch::Event),
+
     /// Change the theme preference.
     Theme(UserTheme),
+
+    /// Change the update check preference.
+    UpdateCheck,
+
+    /// Show update message.
+    UpdateAvailable(UpdateNotification),
 }
 
 /// How the user wants to handle errors with reading the config file.
@@ -67,43 +76,53 @@ impl Framework {
     /// Create a framework for egui.
     pub(crate) fn new(
         size: PhysicalSize<u32>,
-        scale_factor: f64,
+        scale_factor: f32,
         theme: Theme,
         gui: Gui,
         gpu: &Gpu,
+        event_loop_proxy: EventLoopProxy<UserEvent>,
     ) -> Self {
         let width = size.width;
         let height = size.height;
         let font_definitions = create_fonts();
         let style = create_style(theme);
-        let platform = Platform::new(PlatformDescriptor {
-            physical_width: width,
-            physical_height: height,
-            scale_factor,
-            font_definitions,
-            style,
-        });
+
+        let egui_state = egui_winit::State::from_pixels_per_point(scale_factor);
+        let egui_ctx = CtxRef::default();
         let screen_descriptor = ScreenDescriptor {
             physical_width: width,
             physical_height: height,
-            scale_factor: scale_factor as f32,
+            scale_factor,
         };
         let rpass = RenderPass::new(&gpu.device, wgpu::TextureFormat::Bgra8UnormSrgb, 1);
 
-        Self {
-            start_time: Instant::now(),
-            platform,
+        egui_ctx.set_fonts(font_definitions);
+        egui_ctx.set_style(style);
+
+        let mut result = Self {
+            egui_ctx,
+            egui_state,
             screen_descriptor,
             rpass,
             paint_jobs: Vec::new(),
             theme: None,
             gui,
-        }
+            update_checker: None,
+            event_loop_proxy,
+        };
+        result.create_update_checker();
+
+        result
     }
 
     /// Handle input events from the window manager.
-    pub(crate) fn handle_event<T>(&mut self, event: &winit::event::Event<'_, T>) {
-        self.platform.handle_event(event);
+    pub(crate) fn handle_event(&mut self, event: &winit::event::WindowEvent) {
+        self.egui_state.on_event(&self.egui_ctx, event);
+    }
+
+    /// Handle file system change events.
+    pub(crate) fn handle_fs_change(&mut self, event: hotwatch::Event) {
+        self.gui.handle_fs_change(event);
     }
 
     /// Resize egui.
@@ -119,21 +138,18 @@ impl Framework {
 
     /// Prepare egui.
     pub(crate) fn prepare(&mut self, window: &Window) {
-        self.platform
-            .update_time(self.start_time.elapsed().as_secs_f64());
+        update_theme(&mut self.theme, &self.egui_ctx);
 
-        // Begin the egui frame.
-        self.platform.begin_frame();
+        // Run the egui frame and create all paint jobs to prepare for rendering.
+        let raw_input = self.egui_state.take_egui_input(window);
+        let (output, paint_commands) = self.egui_ctx.run(raw_input, |egui_ctx| {
+            // Draw the application GUI.
+            self.gui.ui(egui_ctx, window);
+        });
 
-        // Draw the application GUI.
-        let ctx = self.platform.context();
-        self.update_theme(&ctx);
-        self.gui.ui(&ctx, window);
-
-        // End the egui frame and create all paint jobs to prepare for rendering.
-        // TODO: Handle output.needs_repaint to avoid game-mode continuous redraws.
-        let (_output, paint_commands) = self.platform.end_frame(Some(window));
-        self.paint_jobs = self.platform.context().tessellate(paint_commands);
+        self.egui_state
+            .handle_output(window, &self.egui_ctx, output);
+        self.paint_jobs = self.egui_ctx.tessellate(paint_commands);
     }
 
     /// Render egui.
@@ -145,7 +161,7 @@ impl Framework {
     ) -> Result<(), BackendError> {
         // Upload all resources to the GPU.
         self.rpass
-            .update_texture(&gpu.device, &gpu.queue, &self.platform.context().texture());
+            .update_texture(&gpu.device, &gpu.queue, &self.egui_ctx.font_image());
         self.rpass.update_user_textures(&gpu.device, &gpu.queue);
         self.rpass.update_buffers(
             &gpu.device,
@@ -230,6 +246,38 @@ impl Framework {
         }
     }
 
+    /// Create an update checker.
+    ///
+    /// Adds a warning window if the checker cannot be created.
+    fn create_update_checker(&mut self) {
+        self.update_checker = match UpdateChecker::new(
+            self.event_loop_proxy.clone(),
+            self.gui.config.get_update_check(),
+        ) {
+            Ok(update_checker) => update_checker,
+            Err(err) => {
+                let warn = ShowWarning::new(err, "Error while creating update checker");
+                self.gui.add_warning(warn);
+
+                None
+            }
+        };
+    }
+
+    /// Set the update check frequency.
+    pub(crate) fn recreate_update_check(&mut self) {
+        // Stop the old update checker
+        if let Some(update_check) = self.update_checker.take() {
+            if let Err(err) = update_check.stop(false) {
+                let warn = ShowWarning::new(err, "Error while stopping update checker");
+                self.gui.add_warning(warn);
+            }
+        }
+
+        // Create a new update checker
+        self.create_update_checker();
+    }
+
     /// Try to save the configuration with the current window geometry.
     ///
     /// Returns true on success. When saving fails, the error is shown to the user and `false` is
@@ -274,15 +322,25 @@ impl Framework {
         self.gui.add_error(err);
     }
 
-    /// Configure the theme based on system settings.
-    fn update_theme(&mut self, ctx: &egui::CtxRef) {
-        if let Some(theme) = self.theme.take() {
-            // Set the style
-            ctx.set_style(create_style(theme));
-        }
+    /// Add an update notification to the GUI.
+    pub(crate) fn add_update_notification(&mut self, notification: UpdateNotification) {
+        self.gui.add_update_notification(notification);
     }
 }
 
+/// Configure the theme based on system settings.
+fn update_theme(theme: &mut Option<Theme>, ctx: &egui::CtxRef) {
+    if let Some(theme) = theme.take() {
+        // Set the style
+        ctx.set_style(create_style(theme));
+    }
+}
+
+pub(crate) fn cache_path() -> PathBuf {
+    ProjectDirs::from("org", "KodeWerx", "CarTunes")
+        .map_or_else(|| PathBuf::from("."), |dir| dir.cache_dir().to_path_buf())
+}
+
 /// Get the application configuration path.
 fn config_path() -> PathBuf {
     // If a project directory cannot be found, use the current working directory.
@@ -304,7 +362,7 @@ fn create_fonts() -> egui::FontDefinitions {
         .0;
     fonts
         .font_data
-        .insert("MonoSpace".to_owned(), Cow::from(font));
+        .insert("MonoSpace".to_owned(), egui::FontData::from_owned(font));
 
     let props = FontPropertyBuilder::new().family("sans-serif").build();
     let font = system_fonts::get(&props)
@@ -316,7 +374,7 @@ fn create_fonts() -> egui::FontDefinitions {
         .0;
     fonts
         .font_data
-        .insert("SansSerif".to_owned(), Cow::from(font));
+        .insert("SansSerif".to_owned(), egui::FontData::from_owned(font));
 
     // Set font families
     fonts
