@@ -4,13 +4,16 @@ use self::grid::SetupGrid;
 use crate::config::{Config, UserTheme};
 use crate::framework::UserEvent;
 use crate::setup::{Setup, Setups};
-use crate::str_ext::Ellipsis;
+use crate::str_ext::{Ellipsis, HumanCompare};
+use crate::updates::{UpdateFrequency, UpdateNotification};
 use copypasta::{ClipboardContext, ClipboardProvider};
 use egui::widgets::color_picker::{color_edit_button_srgba, Alpha};
 use egui::{CtxRef, Widget};
+use hotwatch::Hotwatch;
 use std::collections::{HashMap, VecDeque};
 use std::path::Path;
 use std::time::{Duration, Instant};
+use thiserror::Error;
 use winit::event_loop::EventLoopProxy;
 
 mod grid;
@@ -23,6 +26,9 @@ pub(crate) struct Gui {
     /// A tree of `Setups` containing all known setup exports.
     setups: Setups,
 
+    /// Filesystem watcher for changes to any setup exports.
+    hotwatch: Hotwatch,
+
     /// Selected track name.
     selected_track_name: Option<String>,
 
@@ -44,12 +50,18 @@ pub(crate) struct Gui {
     /// Show the "Warning" window.
     warning: bool,
 
+    /// Show the "Update Notification" window.
+    update_notification: bool,
+
     /// Show an error message.
     show_errors: VecDeque<ShowError>,
 
     /// Show a warning message.
     show_warnings: VecDeque<ShowWarning>,
 
+    /// Show an update notification message.
+    show_update_notification: Option<UpdateNotification>,
+
     /// Show a tooltip.
     show_tooltips: HashMap<egui::Id, (String, Instant)>,
 }
@@ -85,6 +97,12 @@ pub(crate) struct ShowWarning {
     context: String,
 }
 
+#[derive(Debug, Error)]
+pub(crate) enum Error {
+    #[error("File system watch error: {0}")]
+    Notify(#[from] hotwatch::Error),
+}
+
 impl Gui {
     /// Create a GUI.
     pub(crate) fn new(
@@ -93,10 +111,16 @@ impl Gui {
         event_loop_proxy: EventLoopProxy<UserEvent>,
         show_errors: VecDeque<ShowError>,
         show_warnings: VecDeque<ShowWarning>,
-    ) -> Self {
-        Self {
+    ) -> Result<Self, Error> {
+        let mut hotwatch = Hotwatch::new()?;
+        let watcher = Self::watch_setups_path(event_loop_proxy.clone());
+
+        hotwatch.watch(config.get_setups_path(), watcher)?;
+
+        Ok(Self {
             config,
             setups,
+            hotwatch,
             selected_track_name: None,
             selected_car_name: None,
             selected_setups: Vec::new(),
@@ -104,10 +128,12 @@ impl Gui {
             about: false,
             preferences: false,
             warning: false,
+            update_notification: false,
             show_errors,
             show_warnings,
+            show_update_notification: None,
             show_tooltips: HashMap::new(),
-        }
+        })
     }
 
     /// Draw the UI using egui.
@@ -119,18 +145,21 @@ impl Gui {
         egui::TopBottomPanel::top("menubar-container").show(ctx, |ui| {
             ui.set_enabled(enabled);
             egui::menu::bar(ui, |ui| {
-                egui::menu::menu(ui, "File", |ui| {
+                ui.menu_button("File", |ui| {
                     ui.set_min_width(200.0);
                     if ui.button("Preferences").clicked() {
+                        ui.close_menu();
                         self.preferences = true;
                     }
                 });
-                egui::menu::menu(ui, "Help", |ui| {
+                ui.menu_button("Help", |ui| {
                     ui.set_min_width(200.0);
                     if ui.button("About CarTunes...").clicked() {
+                        ui.close_menu();
                         self.about = true;
                     }
                     if ui.button("Support CarTunes on Patreon").clicked() {
+                        ui.close_menu();
                         if let Err(err) = webbrowser::open("https://www.patreon.com/blipjoy") {
                             let warning = ShowWarning::new(err, "Unable to open web browser.");
                             self.show_warnings.push_front(warning);
@@ -141,35 +170,69 @@ impl Gui {
         });
 
         // Draw the footer
-        if !self.show_warnings.is_empty() {
+        if self.show_update_notification.is_some() || !self.show_warnings.is_empty() {
             egui::TopBottomPanel::bottom("footer-container").show(ctx, |ui| {
-                let rect = ui
-                    .horizontal(|ui| {
-                        let rect = ui.available_rect_before_wrap();
-                        let size = ui.spacing().interact_size.y;
-                        let center = egui::Vec2::splat(size / 2.0);
-                        let yellow = egui::Color32::from_rgb(210, 210, 40);
-                        let len = self.show_warnings.len();
+                if self.show_update_notification.is_some() {
+                    let rect = ui
+                        .horizontal(|ui| {
+                            let rect = ui.available_rect_before_wrap();
+                            let size = ui.spacing().interact_size.y;
+                            let center = egui::Vec2::splat(size / 2.0);
+                            let green = egui::Color32::from_rgb(40, 210, 40);
 
-                        ui.spacing_mut().item_spacing.x /= 2.0;
+                            ui.spacing_mut().item_spacing.x /= 2.0;
+                            ui.painter()
+                                .circle_filled(rect.min + center, center.x - 3.0, green);
+                            ui.add_space(size);
+                            ui.label("Update available");
+                            ui.add_space(0.0);
+                        })
+                        .response
+                        .rect;
+                    let response = ui.interact(
+                        rect,
+                        egui::Id::new("update-notification-button"),
+                        egui::Sense::click(),
+                    );
+
+                    if response.hovered() {
+                        let hovered = ui.visuals().widgets.hovered;
                         ui.painter()
-                            .circle_filled(rect.min + center, center.x - 3.0, yellow);
-                        ui.add_space(size);
-                        ui.label(format!("{} Warning{}", len, if len > 1 { "s" } else { "" }));
-                        ui.add_space(0.0);
-                    })
-                    .response
-                    .rect;
-                let response =
-                    ui.interact(rect, egui::Id::new("warnings-button"), egui::Sense::click());
-
-                if response.hovered() {
-                    let hovered = ui.visuals().widgets.hovered;
-                    ui.painter()
-                        .rect_stroke(rect, hovered.corner_radius, hovered.bg_stroke);
+                            .rect_stroke(rect, hovered.corner_radius, hovered.bg_stroke);
+                    }
+                    if response.clicked() {
+                        self.update_notification = true;
+                    }
                 }
-                if response.clicked() {
-                    self.warning = true;
+                if !self.show_warnings.is_empty() {
+                    let rect = ui
+                        .horizontal(|ui| {
+                            let rect = ui.available_rect_before_wrap();
+                            let size = ui.spacing().interact_size.y;
+                            let center = egui::Vec2::splat(size / 2.0);
+                            let yellow = egui::Color32::from_rgb(210, 210, 40);
+                            let len = self.show_warnings.len();
+
+                            ui.spacing_mut().item_spacing.x /= 2.0;
+                            ui.painter()
+                                .circle_filled(rect.min + center, center.x - 3.0, yellow);
+                            ui.add_space(size);
+                            ui.label(format!("{} Warning{}", len, if len > 1 { "s" } else { "" }));
+                            ui.add_space(0.0);
+                        })
+                        .response
+                        .rect;
+                    let response =
+                        ui.interact(rect, egui::Id::new("warnings-button"), egui::Sense::click());
+
+                    if response.hovered() {
+                        let hovered = ui.visuals().widgets.hovered;
+                        ui.painter()
+                            .rect_stroke(rect, hovered.corner_radius, hovered.bg_stroke);
+                    }
+                    if response.clicked() {
+                        self.warning = true;
+                    }
                 }
             });
         }
@@ -179,9 +242,11 @@ impl Gui {
             ui.set_enabled(enabled);
 
             // Draw car filters
-            ui.horizontal(|ui| {
-                self.track_selection(ui);
-                self.car_selection(ui);
+            egui::containers::ScrollArea::horizontal().show(ui, |ui| {
+                ui.horizontal(|ui| {
+                    self.track_selection(ui);
+                    self.car_selection(ui);
+                });
             });
 
             // Draw setup filters
@@ -190,7 +255,7 @@ impl Gui {
             let (track_name, car_name, setups) = self.setup_selection(ui, &colors);
             if !setups.is_empty() {
                 // Draw setup properties grid
-                egui::containers::ScrollArea::auto_sized()
+                egui::containers::ScrollArea::both()
                     .id_source(format!("{}{}", track_name, car_name))
                     .show(ui, |ui| {
                         SetupGrid::new(ui, &setups, &colors, diff_colors).show(ui, car_name);
@@ -204,13 +269,107 @@ impl Gui {
         if self.warning {
             self.warning_window(ctx, enabled);
         }
+        self.show_update_notification(ctx, enabled);
+    }
+
+    /// Create a file system watcher.
+    fn watch_setups_path(event_loop_proxy: EventLoopProxy<UserEvent>) -> impl Fn(hotwatch::Event) {
+        move |event| {
+            event_loop_proxy
+                .send_event(UserEvent::FsChange(event))
+                .expect("Event loop must exist");
+        }
+    }
+
+    /// Handle file system change events.
+    ///
+    /// Called by the closure from `Self::watch_setups_path`.
+    pub(crate) fn handle_fs_change(&mut self, event: hotwatch::Event) {
+        use crate::setup::UpdateKind::*;
+
+        // Update the setups tree.
+        let updates = self.setups.update(&event, &self.config);
+        for update in updates {
+            match update {
+                AddedSetup(track_name, car_name, index) => {
+                    if self.selected_track_name.as_ref() == Some(&track_name)
+                        && self.selected_car_name.as_ref() == Some(&car_name)
+                    {
+                        // Update selected setups when a new one is added
+                        for i in self.selected_setups.iter_mut() {
+                            if *i >= index {
+                                *i += 1;
+                            }
+                        }
+                    }
+                }
+                RemovedSetup(track_name, car_name, index) => {
+                    if self.selected_track_name.as_ref() == Some(&track_name)
+                        && self.selected_car_name.as_ref() == Some(&car_name)
+                    {
+                        // Update selected setups when an old one is removed
+                        self.selected_setups.retain(|i| *i != index);
+                        for i in self.selected_setups.iter_mut() {
+                            if *i >= index {
+                                *i -= 1;
+                            }
+                        }
+                    }
+                }
+                RemovedCar(track_name, car_name) => {
+                    if self.selected_track_name.as_ref() == Some(&track_name)
+                        && self.selected_car_name.as_ref() == Some(&car_name)
+                    {
+                        self.selected_car_name = None;
+                        self.selected_setups.clear();
+                    }
+                }
+                RemovedTrack(track_name) => {
+                    if self.selected_track_name.as_ref() == Some(&track_name) {
+                        self.selected_track_name = None;
+                        self.selected_car_name = None;
+                        self.selected_setups.clear();
+                    }
+                }
+            }
+        }
+
+        // Show warning window if necessary.
+        if let hotwatch::Event::Error(error, path) = event {
+            let msg = path.map_or("Error while watching file system".to_string(), |path| {
+                format!("Error while watching path: `{:?}`", path)
+            });
+
+            self.show_warnings.push_front(ShowWarning::new(error, msg));
+        }
     }
 
     /// Update setups export path.
     pub(crate) fn update_setups_path<P: AsRef<Path>>(&mut self, setups_path: P) {
+        if let Err(error) = self.hotwatch.unwatch(self.config.get_setups_path()) {
+            self.show_warnings.push_front(ShowWarning::new(
+                error,
+                format!(
+                    "Unable to stop watching setup exports path for changes: `{:?}`",
+                    self.config.get_setups_path()
+                ),
+            ));
+        }
+
         self.config.update_setups_path(setups_path);
         self.setups = Setups::new(&mut self.show_warnings, &self.config);
         self.clear_filters();
+
+        let watcher = Self::watch_setups_path(self.event_loop_proxy.clone());
+        if let Err(error) = self.hotwatch.watch(self.config.get_setups_path(), watcher) {
+            self.show_warnings.push_front(ShowWarning::new(
+                error,
+                format!(
+                    "Unable to watch setup exports path for changes: `{:?}`",
+                    self.config.get_setups_path()
+                ),
+            ));
+        }
     }
 
     /// Clear track, car, and setup filters.
@@ -231,7 +390,7 @@ impl Gui {
         };
         track_selection.show_ui(ui, |ui| {
             let mut track_names: Vec<_> = self.setups.tracks().keys().collect();
-            track_names.sort_unstable();
+            track_names.sort_unstable_by(|a, b| a.human_compare(b));
 
             for track_name in track_names {
                 let checked = self.selected_track_name.as_ref() == Some(track_name);
@@ -280,7 +439,7 @@ impl Gui {
                         .expect("Invalid track name")
                         .keys()
                         .collect();
-                    car_names.sort_unstable();
+                    car_names.sort_unstable_by(|a, b| a.human_compare(b));
 
                     for car_name in car_names {
                         let checked = self.selected_car_name.as_ref() == Some(car_name);
@@ -315,27 +474,25 @@ impl Gui {
                     output_track_name = track_name.as_str();
                     output_car_name = car_name.as_str();
 
-                    let mut setups: Vec<_> = tracks
+                    let setups = tracks
                         .get(track_name)
                         .expect("Invalid track name")
                         .get(car_name)
-                        .expect("Invalid car name")
-                        .iter()
-                        .collect();
-                    setups.sort_unstable_by(|(a, _), (b, _)| a.partial_cmp(b).unwrap());
+                        .expect("Invalid car name");
 
-                    for (i, (name, _)) in setups.iter().enumerate() {
+                    for (i, info) in setups.iter().enumerate() {
                         let position = selected_setups.iter().position(|&v| v == i);
                         let mut checked = position.is_some();
                         let color = position
-                            .map(|i| colors.iter().cycle().nth(i))
-                            .flatten()
+                            .and_then(|i| colors.iter().cycle().nth(i))
                             .cloned()
                             .unwrap_or_else(|| ui.visuals().text_color());
 
-                        let checkbox = egui::Checkbox::new(&mut checked, name)
-                            .text_color(color)
-                            .ui(ui);
+                        let checkbox = egui::Checkbox::new(
+                            &mut checked,
+                            egui::RichText::new(info.name()).color(color),
+                        )
+                        .ui(ui);
                         if checkbox.clicked() {
                             if checked {
                                 selected_setups.push(i);
@@ -346,7 +503,7 @@ impl Gui {
                     }
 
                     for i in selected_setups {
-                        output.push(&setups[*i].1);
+                        output.push(setups[*i].setup());
                     }
                 }
             }
@@ -394,12 +551,12 @@ impl Gui {
 
                     ui.label("Theme:");
                     egui::ComboBox::from_id_source("theme-preference")
-                        .selected_text(current_theme)
+                        .selected_text(current_theme.to_string())
                         .show_ui(ui, |ui| {
                             let choices = [UserTheme::Auto, UserTheme::Dark, UserTheme::Light];
                             for choice in &choices {
                                 let checked = current_theme == *choice;
-                                let response = ui.selectable_label(checked, format!("{}", choice));
+                                let response = ui.selectable_label(checked, choice.to_string());
                                 if response.clicked() {
                                     self.config.update_theme(*choice);
                                     self.event_loop_proxy
@@ -410,14 +567,47 @@ impl Gui {
                         });
                 });
 
+                // Update check frequency
+                ui.horizontal(|ui| {
+                    let update_check = self.config.get_update_check();
+
+                    ui.label("Update checks:");
+                    egui::ComboBox::from_id_source("update-check-preference")
+                        .selected_text(update_check.to_string())
+                        .show_ui(ui, |ui| {
+                            let choices = [
+                                UpdateFrequency::Never,
+                                UpdateFrequency::Daily,
+                                UpdateFrequency::Weekly,
+                            ];
+                            for choice in &choices {
+                                let checked = update_check == *choice;
+                                let response = ui.selectable_label(checked, choice.to_string());
+                                if response.clicked() {
+                                    self.config.set_update_check(*choice);
+                                    self.event_loop_proxy
+                                        .send_event(UserEvent::UpdateCheck)
+                                        .expect("Event loop must exist");
+                                }
+                            }
+                        });
+                });
+
                 // Setup exports path selection
                 ui.horizontal(|ui| {
                     let setups_path = self.config.get_setups_path();
-                    let label = setups_path.to_string_lossy().ellipsis(50);
+
+                    // Strip Windows path prefixes for display in the GUI
+                    let label = setups_path.to_string_lossy();
+                    let label = label
+                        .strip_prefix(r"\\?\UNC\")
+                        .or_else(|| label.strip_prefix(r"\\?\"))
+                        .or_else(|| label.strip_prefix(r"\??\"))
+                        .unwrap_or(&label)
+                        .ellipsis(50);
 
                     ui.label("Setup exports path:");
-                    if egui::Label::new(label)
-                        .code()
+                    if egui::Label::new(egui::RichText::new(label).code())
                         .sense(egui::Sense::click())
                         .ui(ui)
                         .clicked()
@@ -443,7 +633,7 @@ impl Gui {
                 ui.separator();
                 ui.label("Column colors:");
                 ui.horizontal_wrapped(|ui| {
-                    let colors = self.config.mut_colors();
+                    let colors = self.config.colors_mut();
                     let mut changed = false;
                     let mut to_delete = None;
 
@@ -475,7 +665,7 @@ impl Gui {
 
                 ui.label("Diff colors:");
                 ui.horizontal_wrapped(|ui| {
-                    let colors = self.config.mut_diff_colors();
+                    let colors = self.config.diff_colors_mut();
                     let old_colors = *colors;
 
                     color_edit_button_srgba(ui, &mut colors.0, Alpha::Opaque);
@@ -514,15 +704,18 @@ impl Gui {
                 .show(ctx, |ui| {
                     ui.label(&err.context);
 
-                    egui::ScrollArea::from_max_height(height).show(ui, |ui| {
-                        egui::TextEdit::multiline(&mut err.error.to_string())
-                            .enabled(false)
-                            .text_style(egui::TextStyle::Monospace)
-                            .text_color(red)
-                            .desired_width(width)
-                            .desired_rows(10)
-                            .ui(ui);
-                    });
+                    egui::ScrollArea::vertical()
+                        .max_height(height)
+                        .show(ui, |ui| {
+                            let mut text = err.error.to_string();
+                            let text_edit = egui::TextEdit::multiline(&mut text)
+                                .text_style(egui::TextStyle::Monospace)
+                                .text_color(red)
+                                .desired_width(width)
+                                .desired_rows(10);
+
+                            ui.add_enabled(false, text_edit);
+                        });
 
                     ui.separator();
                     ui.horizontal(|ui| {
@@ -551,11 +744,13 @@ impl Gui {
                                 let action = err.buttons.0.action;
                                 action();
                                 result = false;
-                            } else if egui::Button::new(&err.buttons.1.label)
-                                .text_color(egui::Color32::BLACK)
-                                .fill(red)
-                                .ui(ui)
-                                .clicked()
+                            } else if egui::Button::new(
+                                egui::RichText::new(&err.buttons.1.label)
+                                    .color(egui::Color32::BLACK),
+                            )
+                            .fill(red)
+                            .ui(ui)
+                            .clicked()
                             {
                                 let action = err.buttons.1.action;
                                 action();
@@ -573,6 +768,14 @@ impl Gui {
         }
     }
 
+    /// Add a warning to the GUI.
+    ///
+    /// The new warning will be shown to the user if it is the only one, or else it will wait in a
+    /// queue until older warnings have been acknowledged.
+    pub(crate) fn add_warning(&mut self, warn: ShowWarning) {
+        self.show_warnings.push_front(warn);
+    }
+
     /// Show warning window.
     fn warning_window(&mut self, ctx: &egui::CtxRef, enabled: bool) {
         let mut window_open = self.warning;
@@ -591,15 +794,18 @@ impl Gui {
                     ui.set_enabled(enabled);
                     ui.label(&warning.context);
 
-                    egui::ScrollArea::from_max_height(height).show(ui, |ui| {
-                        egui::TextEdit::multiline(&mut warning.warning.to_string())
-                            .enabled(false)
-                            .text_style(egui::TextStyle::Monospace)
-                            .text_color(yellow)
-                            .desired_width(width)
-                            .desired_rows(10)
-                            .ui(ui);
-                    });
+                    egui::ScrollArea::vertical()
+                        .max_height(height)
+                        .show(ui, |ui| {
+                            let mut text = warning.warning.to_string();
+                            let text_edit = egui::TextEdit::multiline(&mut text)
+                                .text_style(egui::TextStyle::Monospace)
+                                .text_color(yellow)
+                                .desired_width(width)
+                                .desired_rows(10);
+
+                            ui.add_enabled(false, text_edit);
+                        });
 
                     ui.separator();
                     ui.horizontal(|ui| {
@@ -635,6 +841,41 @@ impl Gui {
         self.warning = window_open;
     }
 
+    /// Add a n update notification to the GUI.
+    pub(crate) fn add_update_notification(&mut self, notification: UpdateNotification) {
+        self.show_update_notification = Some(notification);
+    }
+
+    /// Show update notification window.
+    fn show_update_notification(&mut self, ctx: &egui::CtxRef, enabled: bool) {
+        let update_notification = self.show_update_notification.as_ref();
+        if let Some(update_notification) = update_notification {
+            egui::Window::new("New update available")
+                .open(&mut self.update_notification)
+                .collapsible(false)
+                .default_pos((125.0, 125.0))
+                .fixed_size((350.0, 120.0))
+                .show(ctx, |ui| {
+                    let size = ui.spacing().interact_size.y;
+
+                    ui.set_enabled(enabled);
+
+                    ui.label(concat!(
+                        "You are using version: ",
+                        env!("CARGO_PKG_VERSION"),
+                    ));
+                    ui.label(format!("New version: {}", update_notification.version));
+                    ui.add_space(size);
+                    ui.label("Release notes:");
+                    ui.label(&update_notification.release_notes);
+
+                    ui.separator();
+
+                    ui.hyperlink_to("Download update", &update_notification.update_url);
+                });
+        }
+    }
+
     /// Add a tooltip to the GUI.
     ///
     /// The tooltip must be displayed until it expires or this will "leak" tooltips.
@@ -730,10 +971,17 @@ fn get_combo_box_width<'a>(ui: &egui::Ui, choices: impl Iterator<Item = &'a Stri
     let spacing = ui.spacing();
     let default = spacing.interact_size.x + spacing.item_spacing.x + spacing.icon_width;
     choices.fold(default, |width, choice| {
-        let galley = ui
-            .fonts()
-            .layout_no_wrap(egui::TextStyle::Button, choice.to_string());
+        let galley = ui.fonts().layout_no_wrap(
+            choice.to_string(),
+            egui::TextStyle::Button,
+            egui::Color32::TEMPORARY_COLOR,
+        );
 
-        width.max(galley.size.x + spacing.item_spacing.x + spacing.icon_width)
+        width.max(
+            galley.rect.width()
+                + spacing.item_spacing.x
+                + spacing.icon_width
+                + spacing.scroll_bar_width,
+        )
     })
 }
